import time
import logging

LOG = logging.getLogger(__name__)

class Timer(object):
	def __init__(self, name, active = True):
		self.name = name
		self.active = active

	def __enter__(self):
		self.start_tmstmp = time.time_ns()

	def __exit__(self, type, value, traceback):
		if self.active:
			LOG.debug(f'Timer {self.name} finished. Took {round((time.time_ns() - self.start_tmstmp) / 1000000, 3)} ms.')


class Pair(object):
	def __init__(self, base : str, quote : str):
		self.base = base
		self.quote = quote

	def __str__(self):
		return self.base + "_" + self.quote

	def __repr__(self):
		return self.__str__()



import websockets
import json
import logging
import asyncio
from abc import ABC, abstractmethod
from typing import List, Callable, Any

from bitpanda.Pair import Pair
from bitpanda import enums

LOG = logging.getLogger(__name__)

class Subscription(ABC):
	def __init__(self, callbacks = None):
		self.callbacks = callbacks

	@abstractmethod
	def get_channel_name(self) -> str:
		pass

	@abstractmethod
	def get_channel_subscription_message(self) -> dict:
		pass

	async def process_message(self, response : dict) -> None:
		await self.process_callbacks(response)

	async def process_callbacks(self, response : dict) -> None:
		if self.callbacks is not None:
			await asyncio.gather(*[asyncio.create_task(cb(response)) for cb in self.callbacks])

	@staticmethod
	def _get_subscription_instrument_codes(pairs : List[Pair]) -> List[str]:
		return [pair.base + "_" + pair.quote for pair in pairs]

class SubscriptionMgr(object):
	WEB_SOCKET_URI = "wss://streams.exchange.bitpanda.com"

	def __init__(self, subscriptions : List[Subscription], api_key : str, ssl_context = None):
		self.api_key = api_key
		self.ssl_context = ssl_context

		self.subscriptions = subscriptions

	async def run(self) -> None:
		try:
			# main loop ensuring proper reconnection after a graceful connection termination by the remote server
			while True:
				LOG.debug(f"Initiating websocket connection.")
				async with websockets.connect(SubscriptionMgr.WEB_SOCKET_URI, ssl = self.ssl_context) as websocket:
					subscription_message = self._create_subscription_message()
					LOG.debug(f"> {subscription_message}")
					await websocket.send(json.dumps(subscription_message))

					# start processing incoming messages
					while True:
						response = json.loads(await websocket.recv())
						LOG.debug(f"< {response}")

						# subscription positive response
						if response['type'] == "SUBSCRIPTIONS":
							LOG.info(f"Subscription confirmed for channels [" + ",".join([channel["name"] for channel in response["channels"]]) + "]")

						# subscription negative response
						elif response['type'] == "ERROR":
							raise Exception(f"Subscription error. Request [{json.dumps(subscription_message)}] Response [{json.dumps(response)}]")

						# remote termination with an opportunity to reconnect
						elif response["type"] == "CONNECTION_CLOSING":
							LOG.warning(f"Server is performing connection termination with an opportunity to reconnect.")
							break

						# heartbeat message
						elif response["type"] == "HEARTBEAT":
							pass

						# regular message
						else:
							await self.process_message(response)
		except asyncio.CancelledError:
			LOG.warning(f"Websocket requested to be shutdown.")
		except Exception:
			LOG.error(f"Exception occurred. Websocket will be closed.")
			raise

	def _create_subscription_message(self) -> dict:
		return {
			"type": "SUBSCRIBE",
			"channels": [
				dict(subscription.get_channel_subscription_message(), **{"api_token": self.api_key}) for subscription in self.subscriptions
			]
		}

	async def process_message(self, response : dict) -> None:
		for subscription in self.subscriptions:
			if subscription.get_channel_name() == response["channel_name"]:
				await subscription.process_message(response)
				break

class AccountSubscription(Subscription):
	def __init__(self, callbacks : List[Callable[[dict], Any]] = None):
		super().__init__(callbacks)

	def get_channel_name(self):
		return "ACCOUNT_HISTORY"

	def get_channel_subscription_message(self):
		# api_token property is added automatically by SocketMgr
		return {
			"name": self.get_channel_name(),
		}

class PricesSubscription(Subscription):
	def __init__(self, pairs : List[Pair], callbacks : List[Callable[[dict], Any]] = None):
		super().__init__(callbacks)

		self.pairs = pairs

	def get_channel_name(self):
		return "PRICE_TICKS"

	def get_channel_subscription_message(self):
		return {
			"name": self.get_channel_name(),
			"instrument_codes": Subscription._get_subscription_instrument_codes(self.pairs)
		}

class OrderbookSubscription(Subscription):
	def __init__(self, pairs : List[Pair], depth : str, callbacks : List[Callable[[dict], Any]] = None):
		super().__init__(callbacks)

		self.pairs = pairs
		self.depth = depth

	def get_channel_name(self):
		return "ORDER_BOOK"

	def get_channel_subscription_message(self):
		return {
			"name": self.get_channel_name(),
			"depth": self.depth,
			"instrument_codes": Subscription._get_subscription_instrument_codes(self.pairs)
		}

class CandlesticksSubscriptionParams(object):
	def __init__(self, pair : Pair, unit : enums.TimeUnit, period):
		self.pair = pair
		self.unit = unit
		self.period = period

class CandlesticksSubscription(Subscription):
	def __init__(self, subscription_params : List[CandlesticksSubscriptionParams], callbacks : List[Callable[[dict], Any]] = None):
		super().__init__(callbacks)

		self.subscription_params = subscription_params

	def get_channel_name(self):
		return "CANDLESTICKS"

	def get_channel_subscription_message(self):
		return {
			"name": self.get_channel_name(),
			"properties": [{
				"instrument_code": params.pair.base + "_" + params.pair.quote,
				"time_granularity": {
					"unit": params.unit.value,
					"period": params.period
				}
			} for params in self.subscription_params]
		}

class MarketTickerSubscription(Subscription):
	def __init__(self, pairs : List[Pair], callbacks : List[Callable[[dict], Any]] = None):
		super().__init__(callbacks)

		self.pairs = pairs

	def get_channel_name(self):
		return "MARKET_TICKER"

	def get_channel_subscription_message(self):
		return {
			"name": self.get_channel_name(),
			"instrument_codes": Subscription._get_subscription_instrument_codes(self.pairs)
		}



import enum

class RestCallType(enum.Enum):
	GET = enum.auto()
	POST = enum.auto()
	DELETE = enum.auto()

class OrderSide(enum.Enum):
	BUY = "BUY"
	SELL = "SELL"

class TimeInForce(enum.Enum):
	GOOD_TILL_CANCELLED = "GOOD_TILL_CANCELLED"
	IMMEDIATE_OR_CANCELLED = "IMMEDIATE_OR_CANCELLED"

class TimeUnit(enum.Enum):
	MINUTES = "MINUTES"
	HOURS = "HOURS"
	DAYS = "DAYS"
	WEEKS = "WEEKS"
	MONTHS = "MONTHS"



import setuptools

with open("README.md", "r") as fh:
	long_description = fh.read()

setuptools.setup(
	name="bitpanda-aio",
	version="1.0.0",
	author="nardew",
	author_email="bitpanda.aio@gmail.com",
	description="Bitpanda Global Exchange API asynchronous python client",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/nardew/bitpanda-aio",
	packages=setuptools.find_packages(),
	classifiers=[
		"Development Status :: 4 - Beta",
		"Framework :: AsyncIO",
		"Intended Audience :: Developers",
		"Programming Language :: Python :: 3",
		"Programming Language :: Python :: 3.6",
		"Programming Language :: Python :: 3.7",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
		"Topic :: Software Development :: Libraries",
		"Topic :: Software Development :: Libraries :: Python Modules",
		"Typing :: Typed",
	],
	install_requires=[
		'aiohttp==3.6.2',
		'websockets==8.1',
		'pytz==2019.3'
	],
	python_requires='>=3.6',
)



import asyncio
import aiohttp
import ssl
import logging
import datetime
import pytz
import json
from typing import List, Callable, Any

from bitpanda.Pair import Pair
from bitpanda.subscriptions import Subscription, SubscriptionMgr, PricesSubscription, OrderbookSubscription, AccountSubscription, CandlesticksSubscription, \
	MarketTickerSubscription, CandlesticksSubscriptionParams
from bitpanda import enums
from bitpanda.Timer import Timer

LOG = logging.getLogger(__name__)

class BitpandaClient(object):
	REST_API_URI = "https://api.exchange.bitpanda.com/public/v1/"

	def __init__(self, certificate_path : str = None, api_key : str = None, api_trace_log : bool = False) -> None:
		self.api_key = api_key
		self.api_trace_log = api_trace_log

		self.rest_session = None

		self.ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
		self.ssl_context.load_verify_locations(certificate_path)

		self.subscription_sets = []

	async def get_currencies(self) -> dict:
		return await self._create_get("currencies")

	async def get_account_balances(self) -> dict:
		return await self._create_get("account/balances", headers = self._get_header_api_key())

	async def get_account_fees(self) -> dict:
		return await self._create_get("account/fees", headers = self._get_header_api_key())

	async def get_account_orders(self, from_timestamp : datetime.datetime = None, to_timestamp : datetime.datetime = None,
	                             pair : Pair = None, with_cancelled_and_rejected : str = None, with_just_filled_inactive : str = None,
	                             max_page_size : str = None, cursor : str = None) -> dict:
		params = BitpandaClient._clean_request_params({
			"from": from_timestamp,
			"to": to_timestamp,
			"instrument_code": pair,
			"with_cancelled_and_rejected": with_cancelled_and_rejected,
			"with_just_filled_inactive": with_just_filled_inactive,
			"max_page_size": max_page_size,
			"cursor": cursor,
		})

		try:
			params["from"] = params["from"].astimezone(pytz.utc).isoformat()
		except KeyError:
			pass

		try:
			params["to"] = params["to"].astimezone(pytz.utc).isoformat()
		except KeyError:
			pass

		return await self._create_get("account/orders", params = params, headers = self._get_header_api_key())

	async def get_account_order(self, order_id : str) -> dict:
		return await self._create_get("account/orders/" + order_id, headers=self._get_header_api_key())

	async def get_account_order_trades(self, order_id : str) -> dict:
		return await self._create_get("account/orders/" + order_id + "/trades", headers = self._get_header_api_key())

	async def get_account_trades(self, from_timestamp : datetime.datetime = None, to_timestamp : datetime.datetime = None,
	                             pair : Pair = None, max_page_size : str = None, cursor : str = None) -> dict:
		params = BitpandaClient._clean_request_params({
			"from": from_timestamp,
			"to": to_timestamp,
			"instrument_code": pair,
			"max_page_size": max_page_size,
			"cursor": cursor,
		})

		try:
			params["from"] = params["from"].astimezone(pytz.utc).isoformat()
		except KeyError:
			pass

		try:
			params["to"] = params["to"].astimezone(pytz.utc).isoformat()
		except KeyError:
			pass

		return await self._create_get("account/trades", params = params, headers = self._get_header_api_key())

	async def get_account_trade(self, trade_id : str) -> dict:
		return await self._create_get("account/trades/" + trade_id, headers = self._get_header_api_key())

	async def get_account_trading_volume(self) -> dict:
		return await self._create_get("account/trading-volume", headers = self._get_header_api_key())

	async def create_market_order(self, pair : Pair, side : enums.OrderSide, amount : str, client_id : str = None) -> dict:
		data = {
			"instrument_code": str(pair),
			"side": side.value,
			"type": "MARKET",
			"amount": amount
		}

		if client_id is not None:
			data['client_id'] = client_id

		return await self._create_post("account/orders", data = data, headers = self._get_header_api_key())

	async def create_limit_order(self, pair : Pair, side : enums.OrderSide, amount : str, limit_price : str, time_in_force : enums.TimeInForce = None, client_id : str = None) -> dict:
		data = {
			"instrument_code": str(pair),
			"side": side.value,
			"type": "LIMIT",
			"amount": amount,
			"price": limit_price
		}

		if client_id is not None:
			data['client_id'] = client_id

		if time_in_force is not None:
			data['time_in_force'] = time_in_force.value

		return await self._create_post("account/orders", data = data, headers = self._get_header_api_key())

	async def create_stop_limit_order(self, pair : Pair, side : enums.OrderSide, amount : str, limit_price : str, stop_price : str,
	                                  time_in_force : enums.TimeInForce = None, client_id : str = None) -> dict:
		data = {
			"instrument_code": str(pair),
			"side": side.value,
			"type": "STOP",
			"amount": amount,
			"price": limit_price,
			"trigger_price": stop_price
		}

		if client_id is not None:
			data['client_id'] = client_id

		if time_in_force is not None:
			data['time_in_force'] = time_in_force.value

		return await self._create_post("account/orders", data = data, headers = self._get_header_api_key())

	async def delete_account_orders(self, pair : Pair = None) -> dict:
		params = BitpandaClient._clean_request_params({
			"instrument_code": pair,
		})

		return await self._create_delete("account/orders", params = params, headers = self._get_header_api_key())

	async def delete_account_order(self, order_id : str) -> dict:
		return await self._create_delete("account/orders/" + order_id, headers=self._get_header_api_key())

	async def get_candlesticks(self, pair : Pair, unit : enums.TimeUnit, period : str, from_timestamp : datetime.datetime, to_timestamp : datetime.datetime) -> dict:
		params = {
			"unit": unit.value,
			"period": period,
			"from": from_timestamp.astimezone(pytz.utc).isoformat(),
			"to": to_timestamp.astimezone(pytz.utc).isoformat(),
		}

		return await self._create_get("candlesticks/" + str(pair), params = params)

	async def get_instruments(self) -> dict:
		return await self._create_get("instruments")

	async def get_order_book(self, pair : Pair, level : str = None) -> dict:
		params = BitpandaClient._clean_request_params({
			"level": level,
		})

		return await self._create_get("order-book/" + str(pair), params = params)

	async def get_time(self) -> dict:
		return await self._create_get("time")

	def compose_subscriptions(self, subscriptions : List[Subscription]) -> None:
		self.subscription_sets.append(subscriptions)

	async def start_subscriptions(self) -> None:
		if len(self.subscription_sets):
			done, pending = await asyncio.wait(
				[asyncio.create_task(SubscriptionMgr(subscriptions, self.api_key, self.ssl_context).run()) for subscriptions in self.subscription_sets],
				return_when = asyncio.FIRST_EXCEPTION
			)
			for task in done:
				try:
					task.result()
				except Exception as e:
					LOG.exception(f"Unrecoverable exception occurred while processing messages: {e}")
					LOG.info("All websockets scheduled for shutdown")
					for task in pending:
						if not task.cancelled():
							task.cancel()
		else:
			raise Exception("ERROR: There are no subscriptions to be started.")

	async def close(self) -> None:
		session = self._get_rest_session()
		if session is not None:
			await session.close()

	async def _create_get(self, resource : str, params : dict = None, headers : dict = None) -> dict:
		return await self._create_rest_call(enums.RestCallType.GET, resource, None, params, headers)

	async def _create_post(self, resource : str, data : dict = None, params : dict = None, headers : dict = None) -> dict:
		return await self._create_rest_call(enums.RestCallType.POST, resource, data, params, headers)

	async def _create_delete(self, resource : str, params : dict = None, headers : dict = None) -> dict:
		return await self._create_rest_call(enums.RestCallType.DELETE, resource, None, params, headers)

	async def _create_rest_call(self, rest_call_type : enums.RestCallType, resource : str, data : dict = None, params : dict = None, headers : dict = None) -> dict:
		with Timer('RestCall'):
			if rest_call_type == enums.RestCallType.GET:
				rest_call = self._get_rest_session().get(BitpandaClient.REST_API_URI + resource, json = data, params = params, headers = headers, ssl = self.ssl_context)
			elif rest_call_type == enums.RestCallType.POST:
				rest_call = self._get_rest_session().post(BitpandaClient.REST_API_URI + resource, json = data, params = params, headers = headers, ssl = self.ssl_context)
			elif rest_call_type == enums.RestCallType.DELETE:
				rest_call = self._get_rest_session().delete(BitpandaClient.REST_API_URI + resource, json = data, params = params, headers = headers, ssl = self.ssl_context)
			else:
				raise Exception(f"Unsupported REST call type {rest_call_type}.")

			LOG.debug(f"> resource [{resource}], params [{params}], headers [{headers}], data [{data}]")
			async with rest_call as response:
				status_code = response.status
				response_text = await response.text()

				LOG.debug(f"<: status [{status_code}], response [{response_text}]")

				return {
					"status_code": status_code,
					"response": json.loads(response_text)
				}

	def _get_rest_session(self) -> aiohttp.ClientSession:
		if self.rest_session is not None:
			return self.rest_session

		if self.api_trace_log:
			trace_config = aiohttp.TraceConfig()
			trace_config.on_request_start.append(BitpandaClient._on_request_start)
			trace_config.on_request_end.append(BitpandaClient._on_request_end)
			trace_configs = [trace_config]
		else:
			trace_configs = None

		self.rest_session = aiohttp.ClientSession(trace_configs=trace_configs)

		return self.rest_session

	def _get_header_api_key(self):
		header = {
			"Authorization": "Bearer " + self.api_key
		}

		return header

	@staticmethod
	def _clean_request_params(params : dict) -> dict:
		res = {}
		for key, value in params.items():
			if value is not None:
				res[key] = str(value)

		return res

	async def _on_request_start(session, trace_config_ctx, params) -> None:
		LOG.debug(f"> Context: {trace_config_ctx}")
		LOG.debug(f"> Params: {params}")

	async def _on_request_end(session, trace_config_ctx, params) -> None:
		LOG.debug(f"< Context: {trace_config_ctx}")
		LOG.debug(f"< Params: {params}")
